"""
    This file is part of PyTRiP.

    PyTRiP is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    PyTRiP is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with PyTRiP.  If not, see <http://www.gnu.org/licenses/>
"""
import os
import re
import copy
from math import pi, cos, sin
import numpy as np
import string
import matplotlib._cntr as cntr

from pytrip.error import InputError, ModuleNotLoadedError
import pytrip as plib
import pytrip.dos

try:
    from dicom.dataset import Dataset, FileDataset
    from dicom.sequence import Sequence

    _dicom_loaded = True
except:
    _dicom_loaded = False


class VdxCube:
    """
    VdxCube is the master class for dealing with vois structures,
    a vdxcube object contains VoiCube objects which represent a VOI,
    it could be ex a lung or the tumor.
    The VoiCube contains Slices which correnspons to the CT slices,
    and the slices contains contour object, which contains the contour data,
    a slice can contain multiple, since TRiP only support one contour per slice
    for each voi, it is necessary to merge contour

    VdxCube can import both dicom data and TRiP data,
    and export in the those formats.
    """

    def __init__(self, content, cube=None):
        self.vois = []
        self.cube = cube
        self.version = "1.2"

    def read_dicom(self, data, structure_ids=None):
        if "rtss" not in data:
            raise InputError("Input is not a valid rtss structure")
        dcm = data["rtss"]
        self.version = "2.0"
        for i in range(len(dcm.ROIContours)):
            if structure_ids is None or dcm.ROIContours[i].RefdROINumber in structure_ids:
                v = Voi(dcm.RTROIObservations[i].ROIObservationLabel.decode('iso-8859-1'), self.cube)
                v.read_dicom(dcm.RTROIObservations[i], dcm.ROIContours[i])
                self.add_voi(v)
        self.cube.xoffset = 0
        self.cube.yoffset = 0
        self.cube.zoffset = 0
        """shift = min(self.cube.slice_pos)
        for i in range(len(self.cube.slice_pos)):
                self.cube.slice_pos[i] = self.cube.slice_pos[i]-shift"""

    def get_voi_names(self):
        names = []
        for voi in self.vois:
            names.append(voi.name)
        return names

    def __str__(self):
        return self.get_voi_names

    def add_voi(self, voi):
        self.vois.append(voi)

    def get_voi_by_name(self, name):
        for voi in self.vois:
            if voi.name.lower() == name.lower():
                return voi
        raise InputError("Voi doesn't exist")

    def import_vdx(self, path):
        self.read_vdx(path)

    def read(self, path):
        self.read_vdx(path)

    def read_vdx(self, path):
        fp = open(path, "r")
        content = fp.read().split('\n')
        fp.close()
        i = 0
        n = len(content)
        header_full = False
        #        number_of_vois = 0
        while i < n:
            line = content[i]
            if not header_full:
                if re.match("vdx_file_version", line) is not None:
                    self.version = line.split()[1]
                elif re.match("all_indices_zero_based", line) is not None:
                    self.zero_based = True
#                TODO number_of_vois not used
#                elif re.match("number_of_vois", line) is not None:
#                    number_of_vois = int(line.split()[1])
            if re.match("voi", line) is not None:
                v = Voi(line.split()[1], self.cube)
                if self.version == "1.2":
                    if not line.split()[5] == '0':
                        i = v.read_vdx_old(content, i)
                else:
                    i = v.read_vdx(content, i)
                self.add_voi(v)
                header_full = True
            i += 1

    def concat_contour(self):
        for i in range(len(self.vois)):
            self.vois[i].concat_contour()

    def number_of_vois(self):
        return len(self.vois)

    def write_to_voxel(self, path):
        fp = open(path, "wb+")
        fp.write("vdx_file_version %s\n" % self.version)
        fp.write("all_indices_zero_based\n")
        fp.write("number_of_vois %d\n" % self.number_of_vois())
        self.vois = sorted(self.vois, key=lambda voi: voi.type, reverse=True)
        for i in range(len(self.vois)):
            fp.write(self.vois[i].to_voxel_string().encode('ascii', 'ignore'))
        fp.close()

    def write_to_trip(self, path):
        self.concat_contour()
        self.write_to_voxel(path)

    def write(self, path):
        self.write_to_trip(path)

    def create_dicom(self):
        if _dicom_loaded is False:
            raise ModuleNotLoadedError("Dicom")
        meta = Dataset()
        meta.MediaStorageSOPClassUID = '1.2.840.10008.5.1.4.1.1.2'
        meta.MediaStorageSOPInstanceUID = "1.2.3"
        meta.ImplementationClassUID = "1.2.3.4"
        ds = FileDataset("file", {}, file_meta=meta)
        if self.cube is not None:
            ds.PatientsName = self.patient_name
        else:
            ds.PatientsName = ""
        ds.PatientID = "123456"
        ds.PatientsSex = '0'
        ds.PatientsBirthDate = '19010101'
        ds.SpecificCharacterSet = 'ISO_IR 100'
        ds.AccessionNumber = ''
        ds.is_little_endian = True
        ds.is_implicit_VR = True
        ds.SOPClassUID = '1.2.840.10008.5.1.4.1.1.481.3'
        ds.SOPInstanceUID = '1.2.3'  # !!!!!!!!!!
        ds.StudyInstanceUID = '1.2.3'  # !!!!!!!!!!
        ds.SeriesInstanceUID = '1.2.3'  # !!!!!!!!!!
        ds.FrameofReferenceUID = '1.2.3'  # !!!!!!!!!
        ds.SeriesDate = '19010101'  # !!!!!!!!
        ds.ContentDate = '19010101'  # !!!!!!
        ds.StudyDate = '19010101'  # !!!!!!!
        ds.SeriesTime = '000000'  # !!!!!!!!!
        ds.StudyTime = '000000'  # !!!!!!!!!!
        ds.ContentTime = '000000'  # !!!!!!!!!
        ds.StructureSetLabel = ''
        ds.StructureSetDate = ''
        ds.StructureSetTime = ''
        ds.Modality = 'RTSTRUCT'
        roi_label_list = []
        roi_data_list = []
        roi_structure_roi_list = []

        for i in range(self.number_of_vois()):
            roi_label = self.vois[i].create_dicom_label()
            roi_label.ObservationNumber = str(i + 1)
            roi_label.ReferencedROINumber = str(i + 1)
            roi_label.RefdROINumber = str(i + 1)
            roi_contours = self.vois[i].create_dicom_contour_data(i)
            roi_contours.RefdROINumber = str(i + 1)
            roi_contours.ReferencedROINumber = str(i + 1)

            roi_structure_roi = self.vois[i].create_dicom_structure_roi()
            roi_structure_roi.ROINumber = str(i + 1)

            roi_structure_roi_list.append(roi_structure_roi)
            roi_label_list.append(roi_label)
            roi_data_list.append(roi_contours)
        ds.RTROIObservations = Sequence(roi_label_list)
        ds.ROIContours = Sequence(roi_data_list)
        ds.StructureSetROIs = Sequence(roi_structure_roi_list)
        return ds

    def write_dicom(self, path):
        dcm = self.create_dicom()
        dcm.save_as(os.path.join(path, "rtss.dcm"))


def voi_point_cmp(a, b):
    if abs(a[1] - b[1]) < 0.2:
        c = a[0] - b[0]
    else:
        c = a[1] - b[1]
    if c < 0:
        return -1
    else:
        return 1


def create_cube(cube, name, center, width, height, depth):
    v = Voi(name, cube)
    for i in range(0, cube.dimz):
        z = i * cube.slice_distance
        if center[2] - depth / 2 <= z <= center[2] + depth / 2:
            s = Slice(cube)
            points = [
                [center[0] - width / 2, center[1] - height / 2, z], [center[0] + width / 2, center[1] - height / 2, z],
                [center[0] + width / 2, center[1] + height / 2, z], [center[0] - width / 2, center[1] + height / 2, z],
                [center[0] - width / 2, center[1] - height / 2, z]
            ]
            c = Contour(points, cube)
            s.add_contour(c)
            v.add_slice(s)
    return v


def create_voi_from_cube(cube, name):
    v = Voi(name, cube)
    for i in range(cube.dimz):
        x, y = np.meshgrid(np.arange(len(cube.cube[0, 0])), np.arange(len(cube.cube[0])))
        isodose_obj = cntr.Cntr(x, y, cube.cube[i])
        contour = isodose_obj.trace(100)
        s = Slice(cube)
        if not len(contour):
            continue
        points = np.zeros((len(contour[0]), 3))
        points[:, 0:2] = contour[0] * cube.pixel_size

        points[:, 2] = i * cube.slice_distance
        c = Contour(points.tolist(), cube)
        s.add_contour(c)

        v.add_slice(s)
    return v


def create_cylinder(cube, name, center, radius, depth):
    v = Voi(name, cube)
    t = np.linspace(0, 2 * pi, 100)
    p = zip(center[0] + radius * cos(t), center[1] + radius * sin(t))
    for i in range(0, cube.dimz):
        z = i * cube.slice_distance
        if center[2] - depth / 2 <= z <= center[2] + depth / 2:
            s = Slice(cube)
            points = [[x[0], x[1], z] for x in p]
            c = Contour(points, cube)
            s.add_contour(c)
            v.add_slice(s)
    return v


def create_sphere(cube, name, center, radius):
    v = Voi(name, cube)
    t = np.linspace(0, 2 * pi, 100)
    p = zip(cos(t), sin(t))
    for i in range(0, cube.dimz):
        z = i * cube.slice_distance
        if center[2] - radius <= z <= center[2] + radius:
            r = (radius**2 - (z - center[2])**2)**0.5
            s = Slice(cube)
            points = [[center[0] + r * x[0], center[1] + r * x[1], z] for x in p]
            c = Contour(points, cube)
            s.add_contour(c)
            v.add_slice(s)
    return v


class Voi:
    sagital = 2
    coronal = 1

    def __init__(self, name, cube=None):
        self.cube = cube
        self.name = name
        self.is_concated = False
        self.type = 90
        self.slice_z = []
        self.slices = {}
        self.color = [124, 0, 0]
        self.define_colors()

    def create_copy(self, margin=0):
        voi = copy.deepcopy(self)
        if not margin == 0:
            pass
        return voi

    def get_voi_cube(self):
        if hasattr(self, "voi_cube"):
            return self.voi_cube
        self.voi_cube = pytrip.DosCube(self.cube)
        self.voi_cube.load_from_structure(self, 1000)
        return self.voi_cube

    def add_slice(self, slice):
        key = int(slice.get_position() * 100)
        self.slice_z.append(key)
        self.slices[key] = slice

    def get_name(self):
        return self.name

    def calculate_bad_angles(self, voi):
        pass

    def concat_to_3d_polygon(self):
        self.concat_contour()
        data = []
        for slice in self.slices:
            data.extend(self.slices[slice].contour[0].contour)
        self.polygon3d = np.array(data)

    def get_3d_polygon(self):
        if not hasattr(self, "polygon3d"):
            self.concat_to_3d_polygon()
        return self.polygon3d

    def create_point_tree(self):
        points = {}
        self.concat_contour()
        slice_keys = sorted(self.slices.keys())
        for key in slice_keys:
            contour = self.slices[key].contour[0].contour
            p = {}
            for x in contour:
                p[x[0], x[1], x[2]] = []
            points.update(p)
        n_slice = len(slice_keys)
        last_contour = None
        for i, key in enumerate(slice_keys):
            contour = self.slices[key].contour[0].contour
            n_points = len(contour)
            if i < n_slice - 1:
                next_contour = self.slices[slice_keys[i + 1]].contour[0].contour
            else:
                next_contour = None
            for j, point in enumerate(contour):
                j2 = (j + 1) % (n_points - 2)
                point2 = contour[j2]
                points[(point[0], point[1], point[2])].append(point2)
                points[(point2[0], point2[1], point2[2])].append(point)
                if next_contour is not None:
                    point3 = pytrip.res.point.get_nearest_point(point, next_contour)
                    points[(point[0], point[1], point[2])].append(point3)
                    points[(point3[0], point3[1], point3[2])].append(point)
                if last_contour is not None:
                    point4 = pytrip.res.point.get_nearest_point(point, last_contour)
                    if point4 not in points[(point[0], point[1], point[2])]:
                        points[(point[0], point[1], point[2])].append(point4)
                        points[(point4[0], point4[1], point4[2])].append(point)
            last_contour = contour
        self.points = points

    def get_2d_projection_on_basis(self, basis, offset=None):
        a = np.array(basis[0])
        b = np.array(basis[1])
        self.concat_contour()
        bas = np.array([a, b])
        data = self.get_3d_polygon()
        product = np.dot(data, np.transpose(bas))

        compare = (self.cube.pixel_size)
        filtered = plib.filter_points(product, compare / 2)
        filtered = np.array(sorted(filtered, cmp=voi_point_cmp))
        filtered = plib.points_to_contour(filtered)
        product = filtered

        if offset is not None:
            offset_proj = np.array([np.dot(offset, a), np.dot(offset, b)])
            product = product[:] - offset_proj
        return product

    def get_2d_slice(self, plane, depth):
        self.concat_contour()
        points1 = []
        points2 = []
        for key in sorted(self.slice_z):
            slice = self.slices[key]
            if plane is self.sagital:
                point = sorted(
                    plib.slice_on_plane(np.array(slice.contour[0].contour), plane, depth), key=lambda x: x[1])
            elif plane is self.coronal:
                point = sorted(
                    plib.slice_on_plane(np.array(slice.contour[0].contour), plane, depth), key=lambda x: x[0])
            if len(point) > 0:
                points2.append(point[-1])
                if len(point) > 1:
                    points1.append(point[0])
        s = None
        if len(points1) > 0:
            points1.extend(reversed(points2))
            points1.append(points1[0])
            s = Slice()
            s.add_contour(Contour(points1))
        return s

    def define_colors(self):
        self.colors = []
        self.colors.append([0, 0, 255])
        self.colors.append([0, 128, 0])
        self.colors.append([0, 255, 0])
        self.colors.append([255, 0, 0])
        self.colors.append([0, 128, 128])
        self.colors.append([255, 255, 0])

    def calculate_center(self):
        if hasattr(self, "center_pos"):
            return self.center_pos
        self.concat_contour()
        tot_volume = 0
        center_pos = np.array([0, 0, 0])
        for key in self.slices:
            center, area = self.slices[key].calculate_center()
            tot_volume += area
            center_pos += area * center
        self.center_pos = center_pos / tot_volume
        return center_pos / tot_volume

    def get_color(self, i=None):
        if i is None:
            return self.color
        return self.colors[i % len(self.colors)]

    def set_color(self, color):
        self.color = color

    def create_dicom_label(self):
        roi_label = Dataset()
        roi_label.ROIObservationLabel = self.name
        roi_label.RTROIInterpretedType = self.get_roi_type_name(self.type)
        return roi_label

    def create_dicom_structure_roi(self):
        roi = Dataset()
        roi.ROIName = self.name
        return roi

    def create_dicom_contour_data(self, i):
        roi_contours = Dataset()
        contours = []
        for k in self.slices:
            contours.extend(self.slices[k].create_dicom_contours())
        roi_contours.Contours = Sequence(contours)
        roi_contours.ROIDisplayColor = self.get_color(i)

        return roi_contours

    def read_vdx_old(self, content, i):
        line = content[i]
        items = line.split()
        self.name = items[1]
        self.type = int(items[3])
        i += 1
        #        slices = 10000
        while i < len(content):
            line = content[i]
            if re.match("voi", line) is not None:
                break
#            TODO slices never used - does it make sense ?
#            if re.match("#TransversalObjects", line) is not None:
#                slices = int(line.split()[1])
            i += 1
        print(items)
        return i - 1

    def read_vdx(self, content, i):
        line = content[i]
        self.name = string.join(line.split()[1:], ' ')
        number_of_slices = 10000
        i += 1
        while i < len(content):
            line = content[i]
            if re.match("key", line) is not None:
                self.key = line.split()[1]
            elif re.match("type", line) is not None:
                self.type = int(line.split()[1])
            elif re.match("number_of_slices", line) is not None:
                number_of_slices = int(line.split()[1])
            elif re.match("slice", line) is not None:
                s = Slice()
                i = s.read_vdx(content, i)
                key = int((float(s.get_position()) - min(self.cube.slice_pos)) * 100)
                self.slice_z.append(key)
                self.slices[key] = s
            elif re.match("voi", line) is not None:
                break
            elif len(self.slices) >= number_of_slices:
                break
            i += 1
        return i - 1

    def get_roi_type_number(self, type_name):
        if type_name == 'EXTERNAL':
            return 0
        elif type_name == 'AVOIDANCE':
            return 0
        elif type_name == 'ORGAN':
            return 0
        elif type_name == 'GTV':
            return 1
        elif type_name == 'CTV':
            return 1
        else:
            return 0

    def get_roi_type_name(self, type_id):
        if type_id == 10:
            return "EXTERNAL"
        elif type_id == 2:
            return 'AVOIDANCE'
        elif type_id == 1:
            return 'CTV'
        elif type_id == 0:
            return 'other'
        return ''

    def read_dicom(self, info, data):

        if "Contours" not in data.dir() and "ContourSequence" not in data.dir():
            return

        self.type = self.get_roi_type_number(np.typename)
        self.color = data.ROIDisplayColor
        if "Contours" in data.dir():
            contours = data.Contours
        else:
            contours = data.ContourSequence
        for i in range(len(contours)):
            key = int((float(contours[i].ContourData[2]) - min(self.cube.slice_pos)) * 100)
            if key not in self.slices:
                self.slices[key] = Slice(self.cube)
                self.slice_z.append(key)
            self.slices[key].add_dicom_contour(contours[i])

    def get_thickness(self):
        if len(self.slice_z) <= 1:
            return 3
        return abs(float(self.slice_z[1]) - float(self.slice_z[0])) / 100

    def to_voxel_string(self):
        if len(self.slices) is 0:
            return ""

        out = "\n"
        out += "voi %s\n" % (self.name.replace(" ", "_"))
        out += "key empty\n"
        out += "type %s\n" % self.type
        out += "\n"
        out += "contours\n"
        out += "reference_frame\n"
        out += " origin 0.000 0.000 0.000\n"
        out += " point_on_x_axis 1.000 0.000 0.000\n"
        out += " point_on_y_axis 0.000 1.000 0.000\n"
        out += " point_on_z_axis 0.000 0.000 1.000\n"
        out += "number_of_slices %d\n" % self.number_of_slices()
        out += "\n"
        i = 0
        thickness = self.get_thickness()
        for k in self.slice_z:
            sl = self.slices[k]
            pos = sl.get_position()
            out += "slice %d\n" % i
            out += "slice_in_frame %.3f\n" % pos
            out += "thickness %.3f reference " \
                   "start_pos %.3f stop_pos %.3f\n" % \
                   (thickness, pos - 0.5 * thickness, pos + 0.5 * thickness)
            out += "number_of_contours %d\n" % \
                   self.slices[k].number_of_contours()
            out += self.slices[k].to_voxel_string()
            i += 1
        return out

    def get_row_intersections(self, pos):
        slice = self.get_slice_at_pos(pos[2])
        if slice is None:
            return None
        return np.sort(slice.get_intersections(pos))

    def get_slice_at_pos(self, z):
        thickness = self.get_thickness() / 2 * 100
        for key in self.slices.keys():
            key = key
            low = z * 100 - thickness
            high = z * 100 + thickness
            if (low < key < 100 * z) or (high > key >= 100 * z):
                return self.slices[key]
        return None

    def number_of_slices(self):
        return len(self.slices)

    def concat_contour(self):
        if not self.is_concated:
            for k in self.slices.keys():
                self.slices[k].concat_contour()
        self.is_concated = True

    def get_min_max(self):
        temp_min, temp_max = None, None
        if hasattr(self, "temp_min"):
            return self.temp_min, self.temp_max
        for key in self.slices:
            if temp_min is None:
                temp_min, temp_max = self.slices[key].get_min_max()
            else:
                min1, max1 = self.slices[key].get_min_max()
                temp_min = pytrip.res.point.min_list(temp_min, min1)
                temp_max = pytrip.res.point.max_list(temp_max, max1)
        self.temp_min = temp_min
        self.temp_max = temp_max
        return temp_min, temp_max


class Slice:
    def __init__(self, cube=None):
        self.cube = cube
        self.contour = []
        return

    def add_contour(self, contour):
        self.contour.append(contour)

    def add_dicom_contour(self, dcm):
        offset = []
        offset.append(float(self.cube.xoffset))
        offset.append(float(self.cube.yoffset))
        offset.append(float(min(self.cube.slice_pos)))
        self.contour.append(
            Contour(pytrip.res.point.array_to_point_array(
                np.array(
                    dcm.ContourData, dtype=float), offset)))

    def get_position(self):
        if len(self.contour) == 0:
            return None
        return self.contour[0].contour[0][2]

    def get_intersections(self, pos):
        intersections = []
        for c in self.contour:
            intersections.extend(pytrip.res.point.get_x_intersection(pos[1], c.contour))
        return intersections

    def calculate_center(self):
        tot_area = 0
        center_pos = np.array([0, 0, 0])
        for contour in self.contour:
            center, area = contour.calculate_center()
            tot_area += area
            center_pos += area * center
        return center_pos / tot_area, tot_area

    def read_vdx(self, content, i):
        line = content[i]
        number_of_contours = 0
        i += 1
        while i < len(content):
            line = content[i]
            if re.match("slice_in_frame", line) is not None:
                self.slice_in_frame = float(line.split()[1])
            elif re.match("thickness", line) is not None:
                items = line.split()
                self.thickness = float(items[1])
                if len(items) == 7:
                    self.start_pos = float(items[4])
                    self.stop_pos = float(items[6])
                else:
                    self.start_pos = float(items[3])
                    self.stop_pos = float(items[5])

            elif re.match("number_of_contours", line) is not None:
                number_of_contours = int(line.split()[1])
            elif re.match("contour", line) is not None:
                c = Contour([])
                i = c.read_vdx(content, i)
                self.add_contour(c)
            elif re.match("slice", line) is not None:
                break
            elif self.contour >= number_of_contours:
                break
            i += 1
        return i - 1

    def create_dicom_contours(self):
        contour_list = []
        for i in range(len(self.contour)):
            con = Dataset()
            contour = []
            for p in self.contour[i].contour:
                contour.extend([p[0], p[1], p[2]])
            con.ContourData = contour
            con.ContourGeometricType = 'CLOSED_PLANAR'
            con.NumberofContourPoints = self.contour[i].number_of_points()
            contour_list.append(con)
        return contour_list

    def to_voxel_string(self):
        out = ""
        for i in range(len(self.contour)):
            out += "contour %d\n" % i
            out += "internal false\n"
            out += "number_of_points %d\n" % (self.contour[i].number_of_points() + 1)
            out += self.contour[i].to_voxel_string()
            out += "\n"
        return out

    def number_of_contours(self):
        return len(self.contour)

    def concat_contour(self):
        for i in range(len(self.contour) - 1, 0, -1):
            self.contour[0].push(self.contour[i])
            self.contour.pop(i)
        self.contour[0].concat()

    def remove_inner_contours(self):
        for i in range(len(self.contour) - 1, 0, -1):
            self.contour[0].push(self.contour[i])
            self.contour.pop(i)
        self.contour[0].remove_inner_contours()

    def get_min_max(self):
        temp_min, temp_max = self.contour[0].get_min_max()
        for i in range(1, len(self.contour)):
            min1, max1 = self.contour[i].get_min_max()
            temp_min = pytrip.res.point.min_list(temp_min, min1)
            temp_max = pytrip.res.point.max_list(temp_max, max1)
        return temp_min, temp_max


class Contour:
    def __init__(self, contour, cube=None):
        self.cube = cube
        self.children = []
        self.contour = contour

    def push(self, contour):
        for i in range(len(self.children)):
            if self.children[i].contains_contour(contour):
                self.children[i].push(contour)
                return
        self.add_child(contour)

    def calculate_center(self):
        points = self.contour
        points.append(points[-1])
        points = np.array(points)
        dx_dy = np.array([points[i + 1] - points[i] for i in range(len(points) - 1)])
        if abs(points[0, 2] - points[1, 2]) < 0.01:
            area = -sum(points[0:len(points) - 1, 1] * dx_dy[:, 0])
            paths = np.array((dx_dy[:, 0]**2 + dx_dy[:, 1]**2)**0.5)
        elif abs(points[0, 1] - points[1, 1]) < 0.01:
            area = -sum(points[0:len(points) - 1, 2] * dx_dy[:, 0])
            paths = np.array((dx_dy[:, 0]**2 + dx_dy[:, 2]**2)**0.5)
        elif abs(points[0, 0] - points[1, 0]) < 0.01:
            area = -sum(points[0:len(points) - 1, 2] * dx_dy[:, 1])
            paths = np.array((dx_dy[:, 1]**2 + dx_dy[:, 2]**2)**0.5)
        total_path = sum(paths)

        center = np.array([sum(points[0:len(points) - 1, 0] * paths) / total_path,
                           sum(points[0:len(points) - 1:, 1] * paths) / total_path, points[0, 2]])

        return center, area

    def get_min_max(self):
        min_x = np.amin(np.array(self.contour)[:, 0])
        min_y = np.amin(np.array(self.contour)[:, 1])
        min_z = np.amin(np.array(self.contour)[:, 2])

        max_x = np.amax(np.array(self.contour)[:, 0])
        max_y = np.amax(np.array(self.contour)[:, 1])
        max_z = np.amax(np.array(self.contour)[:, 2])
        return [min_x, min_y, min_z], [max_x, max_y, max_z]

    def to_voxel_string(self):
        out = ""
        for i in range(len(self.contour)):
            out += " %.3f %.3f %.3f %.3f %.3f %.3f\n" % (self.contour[i][0], self.contour[i][1], self.contour[i][2], 0,
                                                         0, 0)
        out += " %.3f %.3f %.3f %.3f %.3f %.3f\n" % (self.contour[0][0], self.contour[0][1], self.contour[0][2], 0, 0,
                                                     0)
        return out

    def read_vdx(self, content, i):
        set_point = False
        points = 0
        j = 0
        while i < len(content):
            line = content[i]
            if set_point:
                if j >= points - 1:
                    break
                con_dat = line.split()
                self.contour.append([float(con_dat[0]), float(con_dat[1]), float(con_dat[2])])
                j += 1
            else:
                if re.match("internal_false", line) is not None:
                    self.internal_false = True
                if re.match("number_of_points", line) is not None:
                    points = int(line.split()[1])
                    set_point = True
            i += 1
        return i - 1

    def add_child(self, contour):
        remove_idx = []
        for i in range(len(self.children)):
            if contour.contains_contour(self.children[i]):
                contour.push(self.children[i])
                remove_idx.append(i)
        remove_idx.sort(reverse=True)
        for i in remove_idx:
            self.children.pop(i)
        self.children.append(contour)

    def number_of_points(self):
        return len(self.contour)

    def has_childs(self):
        if len(self.children) > 0:
            return True
        return False

    def print_child(self, level):
        for i in range(len(self.children)):
            print(level * '\t', )
            print(self.children[i].contour)
            self.children[i].print_child(level + 1)

    def contains_contour(self, contour):
        return pytrip.res.point.point_in_polygon(contour.contour[0][0], contour.contour[0][1], self.contour)

    def concat(self):
        for i in range(len(self.children)):
            self.children[i].concat()
        while len(self.children) > 1:
            d = -1
            child = 0
            for i in range(1, len(self.children)):
                i1_temp, i2_temp, d_temp = pytrip.res.point.short_distance_polygon_idx(
                    self.children[0].contour, self.children[i].contour)
                if d == -1 or d_temp < d:
                    d = d_temp
                    child = i
            i1_temp, i2_temp, d_temp = pytrip.res.point.short_distance_polygon_idx(
                self.children[0].contour, self.contour)
            if d_temp < d:
                self.merge(self.children[0])
                self.children.pop(0)
            else:
                self.children[0].merge(self.children[child])
                self.children.pop(child)
        if len(self.children) == 1:
            self.merge(self.children[0])
            self.children.pop(0)

    def remove_inner_contours(self):
        for i in range(len(self.children)):
            self.children[i].children = []

    def merge(self, contour):
        if len(self.contour) == 0:
            self.contour = contour.contour
            return
        i1, i2, d = pytrip.res.point.short_distance_polygon_idx(self.contour, contour.contour)
        con = []
        for i in range(i1 + 1):
            con.append(self.contour[i])
        for i in range(i2, len(contour.contour)):
            con.append(contour.contour[i])
        for i in range(i2 + 1):
            con.append(contour.contour[i])
        for i in range(i1, len(self.contour)):
            con.append(self.contour[i])
        self.contour = con
        return



#! /usr/bin/env python

import sys

from pytrip.utils.rst_read import RstfileRead


def main(args=sys.argv[1:]):
    file = args[0]
    a = RstfileRead(file)
    fout = open("sobp.dat", 'w')
    for i in range(a.submachines):
        b = a.submachine[i]
        for j in range(len(b.xpos)):
            fout.writelines("%-10.6f%-10.2f%-10.2f%-10.2f%-10.4e\n" % (b.energy / 1000.0, b.xpos[j] / 10.0, b.ypos[j] /
                                                                       10.0, b.focus / 10.0, b.particles[j]))
    fout.close()


if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]))



from pytrip import dos
from pytrip import ctx
import argparse
import logging
from numpy import arange, meshgrid, ma
import matplotlib.pyplot as plt
from matplotlib import colors
import os
import sys

logger = logging.getLogger(__name__)


def check_compatible(a, b):
    """ Simple comparison of cubes. if X,Y,Z dims are the same, and
    pixel sizes as well, then they are compatible. (Duck typed)
    """
    eps = 1e-5

    if a.dimx != b.dimx:
        logger.error("DIMX does not match: "+str(a.dimx)+" "+str(b.dimx))
        raise Exception("Cubes don't match, check dimx in header.")

    if a.dimy != b.dimy:
        logger.error("DIMY does not match: "+str(a.dimy)+" "+str(b.dimy))
        raise Exception("Cubes don't match, check dimy in header.")

    if a.dimz != b.dimz:
        logger.error("DIMZ does not match: "+str(a.dimz)+" "+str(b.dimz))
        raise Exception("Cubes don't match, check dimz in header.")

    if (a.pixel_size - b.pixel_size) > eps:
        logger.error("Pixel size does not match: "+str(a.pixel_size)+" "+str(b.pixel_size))
        raise Exception("Cubes don't match, check pixel_size in header.")

    if a.slice_dimension != b.slice_dimension:
        logger.error("Slice dimension does not match: "+str(a.slice_dimension)+" "+str(b.slice_dimension))
        raise Exception("Cubes don't match, check slice_dimension in header.")
    return True


def main(args):
    if args.verbosity == 1:
        logger.basicConfig(level=logging.INFO)
    if args.verbosity > 1:
        logger.basicConfig(level=logging.DEBUG)

    logger.info("Dos file: " + args.dos)

    dosbasename = args.dos.split(".")[-2]
    ctxbasename = args.ctx.split(".")[-2]

    d = dos.DosCube()
    fndos = dosbasename + ".dos"
    logger.info("Reading "+fndos)
    d.read(fndos)
    fname_dos = os.path.splitext(args.dos)[0]

    c = ctx.CtxCube()
    fnctx = ctxbasename + ".ctx"
    logger.info("Reading "+fnctx)
    c.read(fnctx)
    logger.info("CTX Cube shape" + str(c.cube.shape))
    logger.info("DOS Cube shape" + str(d.cube.shape))

    # check cube data are compatible
    if d is not None:
        check_compatible(c, d)

    logger.info("Number of slices: " + str(d.dimz))

    xmin = d.xoffset + (0.5*d.pixel_size)  # convert bin to actual position to center of bin
    ymin = d.yoffset + (0.5*d.pixel_size)
    zmin = d.zoffset + (0.5*d.slice_distance)

    xmax = xmin + d.dimx*d.pixel_size
    ymax = ymin + d.dimy*d.pixel_size
    zmax = zmin + d.dimz*d.slice_distance

    dmax = d.cube.max()
    dmin = d.cube.min()

    cmax = c.cube.max()
    cmin = c.cube.min()

    logger.info("Cube first bin pos: {:10.2f} {:10.2f} {:10.2f} [mm]".format(xmin, ymin, zmin))
    logger.info("Cube last bin pos : {:10.2f} {:10.2f} {:10.2f} [mm]".format(xmax, ymax, zmax))
    logger.info("CTX min, max values: {:d} {:d}".format(cmin, cmax))
    logger.info("DOS min, max values: {:d} {:d}".format(dmin, dmax))

    # ctx_levels = arange(-1010.0,cmax+10,50.0)
    # ctx_levels = arange(cmin+10, cmax+10, 50.0)  # alternative ctx colour map
    # dos_levels = arange(0, dmax+50, 5.0)

    x = arange(xmin, xmax, d.pixel_size)
    y = arange(ymin, ymax, d.pixel_size)
    X, Y = meshgrid(x, y)
    Xmax = X.max()
    Xmin = X.min()
    Ymax = Y.max()
    Ymin = Y.min()

    ctx_cb = None
    dos_cb = None

    sstart = args.sstart
    sstop = args.sstop

    if sstart is None:
        sstart = 0
    if sstop is None:
        sstop = d.dimz

    ax = plt.subplot(111, autoscale_on=False)

    # loop over each slice
    for ids in range(sstart, sstop):  # starts at 0

        fnout = fname_dos+"_{:03d}".format(ids)+".png"
        if args.verbosity == 0:
            print("Write slice number: "+str(ids) + "/"+str(d.dimz))
        if args.verbosity > 0:
            logger.info("Write slice number: "+str(ids) + "/"+str(d.dimz) + " to " + fnout)

        ax.cla()
        # ids = 150

        dos_slice = d.cube[ids, :, :]
        # dos_slice *= 0.1  # convert %% to %
        ctx_slice = c.cube[ids, :, :]

        # min/max per slice:
        # dsmax = dos_slice.max()
        # dsmin = dos_slice.min()
        # csmax = ctx_slice.max()
        # csmin = ctx_slice.min()

        plt.xlabel("ct [mm]")
        plt.ylabel("ct [mm]")
        ax.set_aspect(1.0)
        plt.grid(True)

        # extent tells what the x and y coordinates are,
        # so matplotlib can make propper assignment.
        ctx_im = ax.imshow(ctx_slice,
                           cmap=plt.cm.gray,
                           interpolation='bilinear',
                           origin="lower",
                           extent=[Xmin, Xmax,
                                   Ymin, Ymax])
        if args.HUbar:
            if ctx_cb is None:
                ctx_cb = plt.colorbar(ctx_im, ticks=arange(-1000, 3000, 200),
                                      orientation='horizontal')
                ctx_cb.set_label('HU')

        # ------- add dose wash --------

        cmap1 = plt.cm.jet
        cmap1.set_under("k", alpha=0.0)  # seems to be broken! :-( -- Sacrificed goat, now working - David
        cmap1.set_over("k", alpha=0.0)
        cmap1.set_bad("k", alpha=0.0)  # Sacrificial knife here
        tmpdat = ma.masked_where(dos_slice <= dmin, dos_slice)  # Sacrifical goat

        # Do countours
        # dos_im = plt.contourf(X,Y,dos_slice,
        #                      dos_levels,
        #                      cmap=cmap1,
        #                      antialiased=True,linewidths=None)

        dos_im = ax.imshow(tmpdat,
                           interpolation='bilinear',
                           cmap=cmap1,
                           norm=colors.Normalize(vmin=0, vmax=1200, clip=False),
                           alpha=0.7,
                           origin="lower",
                           extent=[Xmin, Xmax,
                                   Ymin, Ymax])

        if dos_cb is None:
            dos_cb = plt.colorbar(dos_im,
                                  # extend='both',
                                  orientation='vertical',
                                  shrink=0.8)
            dos_cb.set_ticks(arange(0, 1300, 200))
            dos_cb.set_label('Relative dose %%')

        ax.set_xlim(Xmax, Xmin)
        ax.set_ylim(Ymax, Ymin)

        # majorLocator = plt.MultipleLocator(1)
        # majorFormatter = plt.FormatStrFormatter('%d')
        minorLocator = plt.MultipleLocator(1.5)

        ax.xaxis.set_minor_locator(minorLocator)
        plt.savefig("foo"+str(ids)+".png")

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument("dos", help="doscube to be loaded")
    parser.add_argument("ctx", help="ctxcube to be loaded")  # todo: could be optional
    parser.add_argument("-v", "--verbosity", action='count',
                        help="increase output verbosity", default=0)
    parser.add_argument("-f", "--from", type=int, dest='sstart', metavar='N',
                        help="Output from slice number N", default=0)
    parser.add_argument("-t", "--to", type=int, dest='sstop', metavar='M',
                        help="Output up to slice number M")
    parser.add_argument("-H", "--HUbar", dest='HUbar', default=False, action='store_true',
                        help="Add HU colour bar")
    parser.parse_args()
    args = parser.parse_args()

    sys.exit(main(args))



"""
    This file is part of PyTRiP.

    PyTRiP is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    PyTRiP is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with PyTRiP.  If not, see <http://www.gnu.org/licenses/>
"""

# do not check this file for PEP8 compatibility
# flake8: noqa

from pytrip.ctx import CtxCube
from pytrip.dos import DosCube
from pytrip.vdx import VdxCube
from pytrip.paths import DensityCube
from pytrip.rst import Rst
from pytrip.let import LETCube
from pytrip.get_dvh import GetDvh

__all__ = ['CtxCube', 'VdxCube', 'GetDvh', 'DosCube', 'DensityCube', 'LETCube', 'dicomhelper', 'res', 'guiutil', 'Rst']

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions



import os
import setuptools
from pkg_resources import parse_version

import numpy as np


def pip_command_output(pip_args):
    """
    Get output (as a string) from pip command
    :param pip_args: list o pip switches to pass
    :return: string with results
    """
    import sys
    import pip
    from io import StringIO
    # as pip will write to stdout we use some nasty hacks
    # to substitute system stdout with our own
    old_stdout = sys.stdout
    sys.stdout = mystdout = StringIO()
    pip.main(pip_args)
    output = mystdout.getvalue()
    mystdout.truncate(0)
    sys.stdout = old_stdout
    return output


def setup_versioneer():
    """
    Generate (temporarily) versioneer.py file in project root directory
    :return:
    """
    try:
        # assume versioneer.py was generated using "versioneer install" command
        import versioneer
        versioneer.get_version()
    except ImportError:
        # it looks versioneer.py is missing
        # lets assume that versioneer package is installed
        # and versioneer binary is present in $PATH
        import subprocess
        try:
            # call versioneer install to generate versioneer.py
            subprocess.check_output(["versioneer", "install"])
        except OSError:
            # it looks versioneer is missing from $PATH
            # probably versioneer is installed in some user directory

            # query pip for list of files in versioneer package
            # line below is equivalen to putting result of
            #  "pip show -f versioneer" command to string output
            output = pip_command_output(["show", "-f", "versioneer"])

            # now we parse the results
            import os
            # find absolute path where *versioneer package* was installed
            # and store it in main_path
            main_path = [x[len("Location: "):] for x in output.splitlines() if x.startswith("Location")][0]
            # find path relative to main_path where
            # *versioneer binary* was installed
            bin_path = [x[len("  "):] for x in output.splitlines() if x.endswith(os.path.sep + "versioneer")][0]

            # exe_path is absolute path to *versioneer binary*
            exe_path = os.path.join(main_path, bin_path)
            # call versioneer install to generate versioneer.py
            # line below is equivalent to running in terminal
            # "python versioneer install"
            subprocess.check_output(["python", exe_path, "install"])


def clean_cache():
    """
    Python won't realise that new module has appeared in the runtime
    We need to clean the cache of module finders. Hacking again
    :return:
    """
    import importlib
    try:  # Python ver < 3.3
        vermod = importlib.import_module("versioneer")
        globals()["versioneer"] = vermod
    except ImportError:
        importlib.invalidate_caches()


def get_version():
    """
    Get project version (using versioneer)
    :return: string containing version
    """
    setup_versioneer()
    clean_cache()
    import versioneer
    version = versioneer.get_version()
    parsed_version = parse_version(version)
    if '*@' in parsed_version[1]:
        import time
        version += str(int(time.time()))
    return version


def get_cmdclass():
    """
    Get setuptools command class
    :return:
    """
    setup_versioneer()
    clean_cache()
    import versioneer
    return versioneer.get_cmdclass()


with open('README.rst') as readme_file:
    readme = readme_file.read()

extensions = [setuptools.Extension(
    'pytriplib',
    sources=[os.path.join('pytrip', 'lib', 'filter_point.c')],
    extra_compile_args=['-fpic'])]

setuptools.setup(
    name='pytrip98',
    version=get_version(),
    packages=setuptools.find_packages(exclude="tests"),
    url='https://github.com/pytrip/pytrip',
    license='GPL',
    author='Jakob Toftegaard, Niels Bassler, Leszek Grzanka',
    author_email='leszek.grzanka@gmail.com',
    description='PyTRiP',
    long_description=readme + '\n',
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Scientific/Engineering :: Medical Science Apps.',
        'Topic :: Scientific/Engineering :: Physics',

        # OS and env
        'Environment :: Console',
        'Operating System :: POSIX',
        'Operating System :: Unix',
        'Operating System :: MacOS',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: C',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation :: CPython'
    ],
    package_data={'pytrip': ['data/*.dat']},
    install_requires=[
        'matplotlib', 'numpy', 'pydicom', 'scipy'
    ],
    include_dirs=[np.get_include()],
    ext_modules=extensions,
    entry_points={
        'console_scripts': [
            'trip2dicom=pytrip.utils.trip2dicom:main',
            'dicom2trip=pytrip.utils.dicom2trip:main',
            'slicedos=pytrip.utils.slicedos:main',
        ],
    },
    cmdclass=get_cmdclass())



#! /usr/bin/env python
""" convert gd files to xmgrace readable ascii data.
"""

import sys
import os
from scipy import interpolate
from pytrip import __file__


class ReadGd(object):
    '''read file'''

    def __init__(self, filename, _dataset=0):

        if os.path.isfile(filename) is False:
            raise IOError("Could not find file " + filename)

        if _dataset > 2:
            print("DOS: Error- only 0,1,2 OER set available. Got:", _dataset)
        path = os.path.dirname(__file__)
        path_data = (os.path.join(path, "data/OER_barendsen.dat"), os.path.join(path, "data/OER_furusawa_HSG_C12.dat"),
                     os.path.join(path, "data/OER_furusawa_V79_C12.dat"))
        fd = open(path_data[_dataset], 'r')
        lines = fd.readlines()
        fd.close()
        x = [line.split()[0] for line in lines]
        y = [line.split()[1] for line in lines]
        us = interpolate.UnivariateSpline(x, y, s=0.0)

        gd_file = open(filename, 'r')
        gd_lines = gd_file.readlines()
        gd_file.close()
        first = True
        ignore_rest = False
        for line in gd_lines:
            if not (line[0].isdigit()):
                string = "#" + line
                if not first:
                    ignore_rest = True
            else:
                first = False
                if ignore_rest:
                    string = "#" + line
                else:
                    let = line.split()[7]
                    oer = us(let)
                    string = ""
                    for item in line.split():
                        string = string + item + " "
                    string = string + str(oer[0]) + "\n"

            sys.stdout.write(string)


if __name__ == '__main__':
    ReadGd(sys.argv[1], 2)



"""
    This file is part of PyTRiP.

    PyTRiP is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    PyTRiP is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with PyTRiP.  If not, see <http://www.gnu.org/licenses/>
"""
# import os
# import numpy as np
# import pytrip.rst, pytrip.ddd, pytrip.field, time, \
#     pytrip.dos, pytrip.ctx, pytrip.vdx, pytrip.paths
# #import matplotlib.pyplot as plt
# from pytrip.res.point import *
#
# path = "/home/jato/Projects/TRiP/robustness/Carbon_Head/3mm"
# datafiles = os.listdir(path)
# i = 1
# with open(os.path.join(path, "results")) as fp:
#     dat = fp.read().split('\n')
# keys = dat[0:-1:7]
# dvh = dat[4:-1:7]
# qual = dat[5:-1:7]
# data = {}
# quality = {}
# for k, v, q in zip(keys, dvh, qual):
#     data[k] = v
#     quality[k] = q.split()[1]
# output = []
# x = []
# x2 = []
# y = []
# for item in datafiles:
#     if os.path.splitext(item)[1] == ".rst":
#         name = os.path.splitext(item)[0]
#         rs = rst.Rst()
#         rs.load_field(os.path.join(path, item))
#         avg = 0
#         i = 0
#         value = 0
#         # ~ for machine in rs.get_submachines():
#         # ~ i += 1
#         # ~ grid = machine.get_raster_grid()
#         # ~ if len(grid) == 1 or len(grid[0]) == 1:
#         # ~ continue
#         # ~ dat = np.gradient(grid)
#         # ~ norm = (dat[0]**2+dat[0]**2)**0.5
#         # ~ value += np.sum(norm)/np.sum(norm>0)/10000
#         # ~
#         i = 1
#
#         machine = rs.get_submachines()[0]
#         grid = machine.get_raster_grid()
#         dat = np.gradient(grid)
#         norm = (dat[0] ** 2 + dat[0] ** 2) ** 0.5
#         value = np.sum(norm) / np.sum(norm > 0) / 10000
#
#         machine = rs.get_submachines()[1]
#         grid = machine.get_raster_grid()
#         dat = np.gradient(grid)
#         norm = (dat[0] ** 2 + dat[0] ** 2) ** 0.5
#         value += np.sum(norm) / np.sum(norm > 0) / 10000
#
#         machine = rs.get_submachines()[-1]
#         grid = machine.get_raster_grid()
#         dat = np.gradient(grid)
#         norm = (dat[0] ** 2 + dat[0] ** 2) ** 0.5
#         value += np.sum(norm) / np.sum(norm > 0) / 10000
#
#         # ~ machine = rs.get_submachines()[-2]
#         # ~ grid = machine.get_raster_grid()
#         # ~ dat = np.gradient(grid)
#         # ~ norm = (dat[0]**2+dat[0]**2)**0.5
#         # ~ value += np.sum(norm)/np.sum(norm>0)/10000
#         # ~
#
#         y.append(1 - float(data[name]))
#         x.append(value / i)
#
#         x2.append(float(quality[name]))
# fig = plt.figure()
#
# ax = fig.add_subplot(2, 1, 1)
# plt.plot(x, y, '.')
#
# ax = fig.add_subplot(2, 1, 2)
# plt.plot(x2, y, '*')
#
# plt.show()



from pytrip.error import InputError
from pytrip.tripexecuter.pytripobj import pytripObj


class Field(pytripObj):
    def __init__(self, name):
        self.save_fields = ["name", "gantry", "couch", "fwhm", "rasterstep", "doseextension", "contourextension",
                            "zsteps", "projectile", "target"]
        self.name = name
        self.gantry = 0.0
        self.couch = 0.0
        self.fwhm = 5.0
        self.rasterstep = [2, 2]
        self.doseextension = 1.2
        self.contourextension = 0.6
        self.rasterfile = None

        self.zsteps = 1.0
        self.projectile = 'C'
        self.target = []
        self.selected = False

    def get_name(self):
        return self.name

    def set_rasterfile(self, path):
        self.rasterfile = path

    def get_rasterfile(self):
        return self.rasterfile

    def set_name(self, name):
        self.name = name

    def is_selected(self):
        return self.selected

    def toogle_selected(self):
        self.selected = not self.selected

    def get_gantry(self):
        return self.gantry

    def set_gantry(self, angle):
        if type(angle) is str and not len(angle):
            angle = 0
        try:
            angle = (float(angle) + 360) % 360
            if angle < 0 or angle > 360:
                raise Exception()
            self.gantry = angle

        except Exception:
            raise InputError("Gantry angle shoud be a " "number between 0 and 360")

    def get_couch(self):
        return self.couch

    def get_rasterstep(self):
        return self.rasterstep

    def set_rasterstep(self, a, b):
        try:
            a = float(a)
            b = float(b)
            if a < 0 or b < 0:
                raise Exception()
            self.rasterstep = [a, b]
        except Exception:
            raise InputError("Rastersteps should be " "larger than 0 and numbers")

    def set_couch(self, angle):
        if type(angle) is str and not len(angle):
            angle = 0
        try:
            angle = (float(angle) + 360) % 360
            if angle < 0 or angle > 360:
                raise Exception()
            self.couch = angle

        except Exception:
            raise InputError("Couch angle shoud be " "a number between 0 and 360")

    def set_doseextension(self, doseextension):
        try:
            doseextension = float(doseextension)
            if doseextension < 0:
                raise Exception()
            self.doseextension = doseextension
        except Exception:
            raise InputError("Doseextension should be larger 0")

    def get_doseextension(self):
        return self.doseextension

    def set_contourextension(self, contourextension):
        try:
            contourextension = float(contourextension)
            if contourextension < 0:
                raise Exception()
            self.contourextension = contourextension
        except Exception:
            raise InputError("Contourextension should be larger 0")

    def get_contourextension(self):
        return self.contourextension

    def set_zsteps(self, zsteps):
        try:
            zsteps = float(zsteps)
            if zsteps < 0:
                raise Exception()
            self.zsteps = zsteps
        except Exception:
            raise InputError("ZSteps should be larger 0")

    def get_zsteps(self):
        return self.zsteps

    def get_fwhm(self):
        return self.fwhm

    def set_fwhm(self, fwhm):
        try:
            fwhm = float(fwhm)
            if fwhm <= 0:
                raise Exception()
            self.fwhm = fwhm

        except Exception:
            raise InputError("Fwhm shoud be a number and larger than 0")

    def get_projectile(self):
        return self.projectile

    def set_projectile(self, projectile):

        if projectile not in ['H', 'C', 'O', 'Ne']:
            raise InputError("Projectile not allowed")
        self.projectile = projectile

    def get_target(self):
        return self.target

    def set_target(self, target):
        if len(target) is 0:
            self.target = []
            return
        target = target.split(",")
        if len(target) is 3:
            try:
                self.target = [float(target[0]), float(target[1]), float(target[2])]
                return
            except Exception:
                # TODO fix that !
                pass
        raise InputError("Target should be empty " "or in the format x,y,z")



#! /usr/bin/env python

import numpy as np
from matplotlib.pyplot import xlabel, ylabel, title, figure, show


class GetDvh(object):
    """This class produces DVHs."""

    def build_DVH(self, data, _xnorm):
        # data is a flat array

        # number of bins to use:
        _binx = 110
        # _biny = 110 # TODO why not used ?

        # volume
        _vol = len(data)

        self.bins = _vol  # the amount of bins in VOI
        self.bins_dge95 = 0
        self.bins_dlt95 = 0
        self.bins_dgt107 = 0

        print("Cube size:", _vol)

        _DVHx = np.arange(0, _binx, 1)  # last element ist 109
        _DVHy = np.zeros(110)

        # normalize contents:
        data = 100.0 * data / float(_xnorm)

        # print("DVH: shape", shape(data)

        for _element in data:
            #           print("DVH: element shalpe",shape(_element)
            if _element >= 95:
                self.bins_dge95 += 1
            else:
                self.bins_dlt95 += 1
            if _element > 107.0:
                self.bins_dgt107 += 1
                if _element <= 109.0:  # prevent exceed
                    for iii in range(int(round(_element))):
                        _DVHy[iii] += 1

                    # normalize y-scale
        _DVHy = 100.0 * _DVHy / float(_vol)

        self.bins_95_107 = self.bins_dge95 - self.bins_dgt107
        self.percent_95_107 = 100 * self.bins_95_107 / self.bins
        self.percent_gt107 = 100 * self.bins_dgt107 / self.bins
        self.percent_lt95 = 100 * self.bins_dlt95 / self.bins

        return [_DVHx, _DVHy]

    def __init__(self, dos, _xnorm, vdx, _cv):

        # generate mask, if not already done so
        vdx._calc_voi(_cv)

        # extract VOI array containing the data
        _mask_indicies = np.nonzero(vdx.mask[_cv].flatten())
        print("DVH: vdx mask:", np.shape(vdx.mask[_cv]))
        print("DVH: mask ind:", np.shape(_mask_indicies))

        print("DVH: dos shape:", np.shape(dos.cube.flatten()))
        print("DVH: voi shape:", np.shape(vdx.mask[_cv].flatten()))

        # TODO why this zero below ?
        self.array = np.take(dos.cube.flatten(), _mask_indicies)[0]
        print("DVH: array shape:", np.shape(self.array))
        # print self.array

        self.DVH = self.build_DVH(self.array, _xnorm)
        self.name = vdx.voi[_cv].name

        self.voxel_size = vdx.header.pixel_size * vdx.header.pixel_size * vdx.header.slice_distance
        self.volume = self.bins * self.voxel_size
        self.volume_cm3 = self.volume * 0.001
        self.min = self.array.min()
        self.max = self.array.max()
        self.mean = self.array.mean()
        self.std = self.array.std()

    def analyze(self):
        print("--------------------------------------------------------------")
        print("                          VOI Analysis")
        print(" Name:", self.name)
        print(" VOI Volume: %.3f cm3" % self.volume_cm3)
        print(" in", self.bins, "bins.")
        print(" Min / Max Dose: %.2f / %.2f" % (self.min, self.max))
        print(" Mean / Stdev  : %.2f / %.2f" % (self.mean, self.std))
        print(" Vol: 95 %% <= D <= 107 %% : %.1f %% in %i bins" % (self.percent_95_107, self.bins_95_107))
        print(" Vol: D > 107 %%          : %.1f %% in %i bins" % (self.percent_gt107, self.bins_dgt107))
        print(" Vol: D < 95 %%           : %.1f %% in %i bins" % (self.percent_lt95, self.bins_dlt95))
        print("--------------------------------------------------------------")

    def plot(self):

        fig = figure()
        a1 = fig.add_subplot(111, autoscale_on=False)
        title(self.name)
        xlabel("Dose %")
        ylabel("Vol %")
        a1.set_xlim(0, 110)
        a1.set_ylim(0, 110)
        a1.plot(self.DVH[0], self.DVH[1])
        # leg = a1.legend(self.name)
        show()



"""
    This file is part of PyTRiP.

    PyTRiP is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    PyTRiP is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with PyTRiP.  If not, see <http://www.gnu.org/licenses/>
"""
from math import log, sqrt, pi

import numpy as np

from pytrip.res.point import angles_from_trip, max_list, min_list
from pytrip.res.point import get_basis_from_angles
import pytrip as plib


def compare_raster_point(a, b):
    if a[1] is b[1]:
        return int(a[0] - b[0])
    return int(a[1] - b[1])


class Field:
    def __init__(self, ddd):
        self.ddd = ddd

    def get_cube_basis(self):
        return get_basis_from_angles(self.gantry_angle, self.couch_angle)

    def load_from_raster_points(self, rst):
        self.rst = rst
        self.bolus = rst.bolus
        self.couch_angle = float(rst.couchangle)
        self.gantry_angle = float(rst.gantryangle)
        self.gantry_angle, self.couch_angle = \
            angles_from_trip(self.gantry_angle, self.couch_angle)
        self.subfields = []
        field_size = None
        for submachine in self.rst.get_submachines():
            sub = SubField(submachine, self.ddd, rst)
            self.subfields.append(sub)
            size = sub.get_size()
            if field_size is None:
                field_size = size
            else:
                field_size[0:5:2] = max_list(size[0:5:2], field_size[0:5:2])
                field_size[1:6:2] = min_list(size[1:6:2], field_size[1:6:2])
        self.field_size = field_size
        self.get_merged_raster_points()

    def get_merged_raster_points(self):
        if hasattr(self, "raster_matrix"):
            return self.raster_matrix
        size = None
        for sub in self.rst.get_submachines():
            tmp = sub.get_raster_min_max()
            if size is None:
                size = tmp
            else:
                size[0:3:2] = min_list(size[0:3:2], tmp[0:3:2])
                size[1:4:2] = max_list(size[1:4:2], tmp[1:4:2])
        matrixs = []
        for sub in self.subfields:
            matrixs.append(sub.get_merge_raster_points(size))
        self.raster_matrix = matrixs
        return np.array(matrixs)

    def get_energy_list(self):
        energy_list = []
        for sub in self.rst.get_submachines():
            energy_list.append(sub.energy)
        return energy_list

    def get_ddd_list(self):
        if hasattr(self, "ddd_list"):
            return self.ddd_list
        e_list = self.get_energy_list()
        self.ddd_list = self.ddd.get_ddd_grid(e_list, 1000)
        return self.ddd_list

    def get_max_dist(self):
        m = 0
        for submachine in self.rst.get_submachines():
            m = max(self.ddd.get_dist(submachine.energy), m)
        self.max_dist = m
        return m


class SubField:
    def __init__(self, submachine, ddd, rst, resolution=0.5):
        self.submachine = submachine
        self.sigma = self.submachine.focus
        self.energy = self.submachine.energy
        self.rst = rst
        self.ddd = ddd

    def get_lateral(self, resolution=0.5):
        if hasattr(self, "lateral") and resolution == resolution:
            return self.lateral
        max_dist = self.get_max_dist()
        dim = np.ceil(np.absolute(max_dist / resolution))
        max_dist = dim * resolution
        self.resolution = resolution
        a = np.meshgrid(np.linspace(0, max_dist, dim), np.linspace(0, max_dist, dim))
        r = (a[0]**2 + a[1]**2)**0.5
        sigma = self.sigma / ((8 * log(2))**0.5)
        lateral = 1 / ((2 * pi * sigma**2)**0.5) * np.exp(-(r**2) / (2 * sigma**2))
        tot_lat = np.zeros((2 * dim - 1, 2 * dim - 1))

        tot_lat[dim - 1:2 * dim - 1, dim - 1:2 * dim - 1] = lateral
        tot_lat[dim - 1:2 * dim - 1, 0:dim - 1] = np.rot90(lateral, 3)[:, 0:dim - 1]
        tot_lat[0:dim - 1, 0:dim - 1] = np.rot90(lateral, 2)[0:dim - 1, 0:dim - 1]
        tot_lat[0:dim - 1, dim - 1:2 * dim - 1] = np.rot90(lateral)[0:dim - 1, :]

        self.lateral = tot_lat
        return self.lateral

    def get_size(self):
        beamsize = self.get_max_dist()
        depth = self.ddd.get_dist(self.submachine.energy)
        lateral = self.submachine.get_raster_min_max()
        lateral[0] -= beamsize
        lateral[1] += beamsize
        lateral[2] -= beamsize
        lateral[3] += beamsize
        lateral.extend([0, depth])
        return lateral

    def get_merge_raster_points(self, size):
        points = np.array(self.get_raster_matrixs(size))
        dim = [len(points[0]), len(points)]

        points = np.reshape(points, (dim[0] * dim[1], 3))
        sigma = self.sigma / ((8 * log(2))**0.5)
        self.points = plib.merge_raster_grid(np.array(points), sigma)
        self.points = np.reshape(self.points, (dim[1], dim[0], 3))
        return self.points

    def get_raster_matrixs(self, size):
        points = sorted(self.submachine.get_raster_points(), cmp=compare_raster_point)
        step = self.submachine.stepsize
        margin = 5
        mat = [
            [[x, y, 0]
             for x in np.linspace(size[0] - margin * step[0], size[1] + margin * step[0], (size[1] - size[0]
                                                                                           ) / step[0] + 1 + 2 * margin)
             ]
            for y in np.linspace(size[2] - margin * step[1], size[3] + margin * step[1], (size[3] - size[2]
                                                                                          ) / step[1] + 1 + 2 * margin)
        ]
        for p in points:
            i = int(p[1] / step[1] - size[2] / step[1]) + margin
            j = int(p[0] / step[0] - size[0] / step[0]) + margin
            mat[i][j] = p
        return mat

    def get_subfield_cube(self, resolution=0.5):

        size = self.get_size()
        # ~ doesn't allow minimum values greater than zero
        dim = np.ceil(np.absolute(np.array(size) / resolution))
        field = np.zeros((dim[5] + dim[4], dim[3] + dim[2] - 1, dim[1] + dim[0] - 1))
        lateral = self.get_lateral(resolution)
        ddd = self.ddd.get_ddd_by_energy(self.energy, np.linspace(0, (dim[5]) * resolution, dim[5] * 10))
        ddd = np.sum(np.reshape(ddd, (dim[5], -1)), axis=1) / 10

        raster_dim = [len(lateral[0]), len(lateral), len(ddd)]
        zero = np.ceil(np.array(self.zero) / resolution)
        raster_field = np.array([x * lateral for x in ddd])
        raster = self.submachine.get_raster_points()
        for point in raster:
            idx = np.array([point[0] / resolution, point[1] / resolution])
            raster_center = idx + zero
            start = np.array(
                [raster_center[0] - (raster_dim[0] + 1) / 2, raster_center[1] - (raster_dim[1] + 1) / 2, 0], dtype=int)
            end = np.array(
                [raster_center[0] + (raster_dim[0] - 1) / 2, raster_center[1] + (raster_dim[1] - 1) / 2, raster_dim[2]],
                dtype=int)
            field[start[2]:end[2], start[1]:end[1], start[0]:end[0]] += point[2] * raster_field
        return field, zero

    def get_max_dist(self):
        max_r = self.sigma * sqrt(log(100) / log(2))
        return max_r
